@@ -15,17 +15,14 @@
 package cmd
 
 import (
-	"bytes"
-	"io"
+	"cmp"
 	"os"
 	"path/filepath"
 	"regexp"
-	"strconv"
 	"strings"
 
 	"github.com/spf13/pflag"
 	"golang.org/x/text/language"
-	"golang.org/x/text/message"
 
 	"cuelang.org/go/cue"
 	"cuelang.org/go/cue/ast"
@@ -35,10 +32,9 @@ import (
 	"cuelang.org/go/cue/parser"
 	"cuelang.org/go/cue/token"
 	"cuelang.org/go/internal"
-	"cuelang.org/go/internal/core/adt"
+	"cuelang.org/go/internal/cuedebug"
 	"cuelang.org/go/internal/encoding"
 	"cuelang.org/go/internal/filetypes"
-	"cuelang.org/go/internal/value"
 )
 
 var requestedVersion = os.Getenv("CUE_SYNTAX_OVERRIDE")
@@ -62,8 +58,8 @@ func defaultConfig() (*config, error) {
 					parser.FromVersion(version),
 					parser.ParseComments,
 				}
-				// TODO: consolidate all options into a single CUE_DEBUG variable.
-				if os.Getenv("CUE_DEBUG_PARSER_TRACE") != "" {
+				cuedebug.Init()
+				if cuedebug.Flags.ParserTrace {
 					options = append(options, parser.Trace)
 				}
 				return parser.ParseFile(name, src, options...)
@@ -73,44 +69,12 @@ func defaultConfig() (*config, error) {
 	}, nil
 }
 
-var inTest = false
-
 func getLang() language.Tag {
-	loc := os.Getenv("LC_ALL")
-	if loc == "" {
-		loc = os.Getenv("LANG")
-	}
+	loc := cmp.Or(os.Getenv("LC_ALL"), os.Getenv("LANG"))
 	loc, _, _ = strings.Cut(loc, ".")
 	return language.Make(loc)
 }
 
-func exitOnErr(cmd *Command, err error, fatal bool) {
-	if err == nil {
-		return
-	}
-
-	// Link x/text as our localizer.
-	p := message.NewPrinter(getLang())
-	format := func(w io.Writer, format string, args ...interface{}) {
-		p.Fprintf(w, format, args...)
-	}
-
-	cwd, _ := os.Getwd()
-
-	w := &bytes.Buffer{}
-	errors.Print(w, err, &errors.Config{
-		Format:  format,
-		Cwd:     cwd,
-		ToSlash: inTest,
-	})
-
-	b := w.Bytes()
-	_, _ = cmd.Stderr().Write(b)
-	if fatal {
-		exit()
-	}
-}
-
 func loadFromArgs(args []string, cfg *load.Config) []*build.Instance {
 	binst := load.Instances(args, cfg)
 	if len(binst) == 0 {
@@ -169,17 +133,25 @@ func (b *buildPlan) instances() iterator {
 	case len(b.orphaned) > 0:
 		i = newStreamingIterator(b)
 	case len(b.insts) > 0:
+		insts, err := buildInstances(b.cmd, b.insts, false)
 		i = &instanceIterator{
 			inst: b.instance,
-			a:    buildInstances(b.cmd, b.insts, false),
+			a:    insts,
+			e:    err,
 			i:    -1,
 		}
-	default:
+	case b.instance != nil:
 		i = &instanceIterator{
 			a: []*instance{b.instance},
 			i: -1,
 		}
 		b.instance = nil
+	default:
+		// No instances; return an iterator with zero values.
+		// This can happen when the input is an empty jsonl file, for example.
+		// Zero iteration may not make much sense in some scenarios like export,
+		// but an empty jsonl file is valid, so doing nothing seems reasonable.
+		i = &instanceIterator{}
 	}
 	if len(b.expressions) > 0 {
 		return &expressionIter{
@@ -365,7 +337,7 @@ func (i *expressionIter) value() cue.Value {
 }
 
 type config struct {
-	outMode filetypes.Mode
+	mode filetypes.Mode
 
 	fileFilter     string
 	reFile         *regexp.Regexp
@@ -411,7 +383,7 @@ func newBuildPlan(cmd *Command, cfg *config) (p *buildPlan, err error) {
 	}
 	cfg.reFile = re
 
-	setTags(cfg.loadCfg, cmd.Flags(), cmd.Parent().Flags())
+	setTags(cfg.loadCfg, cmd.Flags())
 
 	return p, nil
 }
@@ -420,16 +392,11 @@ func (p *buildPlan) matchFile(file string) bool {
 	return p.cfg.reFile.MatchString(file)
 }
 
-func setTags(cfg *load.Config, flags ...*pflag.FlagSet) {
-	// setTags usually receives a subcommmand's flags, plus the parent (root) command's flags,
-	// so that we can support "cue cmd -t foo=bar mycmd" as well as the short form "cue -t foo=bar mycmd".
-	// TODO: once we remove "cue cmd" short-hand support, go back to a single FlagSet parameter.
-	for _, f := range flags {
-		tags, _ := f.GetStringArray(string(flagInject))
-		cfg.Tags = append(cfg.Tags, tags...)
-		if b, _ := f.GetBool(string(flagInjectVars)); b {
-			cfg.TagVars = load.DefaultTagVars()
-		}
+func setTags(cfg *load.Config, flags *pflag.FlagSet) {
+	tags, _ := flags.GetStringArray(string(flagInject))
+	cfg.Tags = append(cfg.Tags, tags...)
+	if b, _ := flags.GetBool(string(flagInjectVars)); b {
+		cfg.TagVars = load.DefaultTagVars()
 	}
 }
 
@@ -440,7 +407,7 @@ type decoderInfo struct {
 
 func (d *decoderInfo) dec(b *buildPlan) *encoding.Decoder {
 	if d.d == nil {
-		d.d = encoding.NewDecoder(d.file, b.encConfig)
+		d.d = encoding.NewDecoder(b.cmd.ctx, d.file, b.encConfig)
 	}
 	return d.d
 }
@@ -463,7 +430,7 @@ func (p *buildPlan) getDecoders(b *build.Instance) (schemas, values []*decoderIn
 			f.Interpretation = p.cfg.interpretation
 		}
 		switch f.Encoding {
-		case build.Protobuf, build.YAML, build.JSON, build.JSONL,
+		case build.Protobuf, build.YAML, build.TOML, build.XML, build.JSON, build.JSONL,
 			build.Text, build.Binary:
 			if f.Interpretation == build.ProtobufJSON {
 				// Need a schema.
@@ -488,9 +455,9 @@ func (p *buildPlan) getDecoders(b *build.Instance) (schemas, values []*decoderIn
 		if b.Module != "" {
 			c.ProtoPath = append(c.ProtoPath, b.Root)
 		}
-		d := encoding.NewDecoder(f, &c)
+		d := encoding.NewDecoder(p.cmd.ctx, f, &c)
 
-		fi, err := filetypes.FromFile(f, p.cfg.outMode)
+		fi, err := filetypes.FromFile(f, p.cfg.mode)
 		if err != nil {
 			return schemas, values, err
 		}
@@ -536,8 +503,6 @@ func parseArgs(cmd *Command, args []string, cfg *config) (p *buildPlan, err erro
 		return nil, err
 	}
 
-	adt.DebugSort, _ = strconv.Atoi(os.Getenv("CUE_DEBUG_SORT_ARCS"))
-
 	builds := loadFromArgs(args, cfg.loadCfg)
 	if builds == nil {
 		return nil, errors.Newf(token.NoPos, "invalid args")
@@ -549,7 +514,7 @@ func parseArgs(cmd *Command, args []string, cfg *config) (p *buildPlan, err erro
 
 	for _, b := range builds {
 		if b.Err != nil {
-			return nil, b.Err
+			return nil, suggestModCommand(b.Err)
 		}
 		switch {
 		case !b.User:
@@ -626,11 +591,14 @@ func parseArgs(cmd *Command, args []string, cfg *config) (p *buildPlan, err erro
 			// TODO: ignore errors here for now until reporting of concreteness
 			// of errors is correct.
 			// See https://github.com/cue-lang/cue/issues/1483.
-			inst := buildInstances(
+			insts, err := buildInstances(
 				p.cmd,
 				[]*build.Instance{schema},
-				true)[0]
-
+				true)
+			if err != nil {
+				return nil, err
+			}
+			inst := insts[0]
 			if err := inst.err; err != nil {
 				return nil, err
 			}
@@ -680,31 +648,46 @@ func parseArgs(cmd *Command, args []string, cfg *config) (p *buildPlan, err erro
 func (b *buildPlan) parseFlags() (err error) {
 	b.mergeData = !b.cfg.noMerge && flagMerge.Bool(b.cmd)
 
-	out := flagOut.String(b.cmd)
-	outFile := flagOutFile.String(b.cmd)
-
-	if strings.Contains(out, ":") && strings.Contains(outFile, ":") {
-		return errors.Newf(token.NoPos,
-			"cannot specify qualifier in both --out and --outfile")
-	}
-	if outFile == "" {
-		outFile = "-"
-	}
-	if out != "" {
-		outFile = out + ":" + outFile
-	}
-	b.outFile, err = filetypes.ParseFile(outFile, b.cfg.outMode)
-	if err != nil {
-		return err
+	b.encConfig = &encoding.Config{
+		Mode:      b.cfg.mode,
+		Stdin:     b.cmd.InOrStdin(),
+		Stdout:    b.cmd.OutOrStdout(),
+		ProtoPath: flagProtoPath.StringArray(b.cmd),
+		AllErrors: flagAllErrors.Bool(b.cmd),
+		PkgName:   flagPackage.String(b.cmd),
+		Strict:    flagStrict.Bool(b.cmd),
 	}
 
-	for _, e := range flagExpression.StringArray(b.cmd) {
-		expr, err := parser.ParseExpr("--expression", e)
+	// For commands with an output mode, like `cue export` or `cue def`.
+	if b.cfg.mode != filetypes.Input {
+		out := flagOut.String(b.cmd)
+		outFile := flagOutFile.String(b.cmd)
+
+		if strings.Contains(out, ":") && strings.Contains(outFile, ":") {
+			return errors.Newf(token.NoPos,
+				"cannot specify qualifier in both --out and --outfile")
+		}
+		if outFile == "" {
+			outFile = "-"
+		}
+		if out != "" {
+			outFile = out + ":" + outFile
+		}
+		b.outFile, err = filetypes.ParseFile(outFile, b.cfg.mode)
 		if err != nil {
 			return err
 		}
-		b.expressions = append(b.expressions, expr)
+
+		for _, e := range flagExpression.StringArray(b.cmd) {
+			expr, err := parser.ParseExpr("--expression", e)
+			if err != nil {
+				return err
+			}
+			b.expressions = append(b.expressions, expr)
+		}
+		b.encConfig.Force = flagForce.Bool(b.cmd)
 	}
+
 	if s := flagSchema.String(b.cmd); s != "" {
 		b.schema, err = parser.ParseExpr("--schema", s)
 		if err != nil {
@@ -715,27 +698,24 @@ func (b *buildPlan) parseFlags() (err error) {
 		// Set a default file filter to only include json and yaml files
 		b.cfg.fileFilter = s
 	}
-	b.encConfig = &encoding.Config{
-		Force:         flagForce.Bool(b.cmd),
-		Mode:          b.cfg.outMode,
-		Stdin:         b.cmd.InOrStdin(),
-		Stdout:        b.cmd.OutOrStdout(),
-		ProtoPath:     flagProtoPath.StringArray(b.cmd),
-		AllErrors:     flagAllErrors.Bool(b.cmd),
-		PkgName:       flagPackage.String(b.cmd),
-		Strict:        flagStrict.Bool(b.cmd),
-		InlineImports: flagInlineImports.Bool(b.cmd),
-		EscapeHTML:    flagEscape.Bool(b.cmd),
+	// These flags exist only in specific output modes.
+	switch b.cfg.mode {
+	case filetypes.Export:
+		b.encConfig.EscapeHTML = flagEscape.Bool(b.cmd)
+	case filetypes.Def:
+		b.encConfig.InlineImports = flagInlineImports.Bool(b.cmd)
 	}
 	return nil
 }
 
-func buildInstances(cmd *Command, binst []*build.Instance, ignoreErrors bool) []*instance {
+func buildInstances(cmd *Command, binst []*build.Instance, ignoreErrors bool) ([]*instance, error) {
 	// TODO:
 	// If there are no files and User is true, then use those?
 	// Always use all files in user mode?
 	instances, err := cmd.ctx.BuildInstances(binst)
-	exitOnErr(cmd, err, true)
+	if err != nil {
+		return nil, err
+	}
 
 	insts := make([]*instance, len(instances))
 	for i, v := range instances {
@@ -749,19 +729,32 @@ func buildInstances(cmd *Command, binst []*build.Instance, ignoreErrors bool) []
 	// TODO: remove ignoreErrors flag and always return here, leaving it up to
 	// clients to check for errors down the road.
 	if ignoreErrors || flagIgnore.Bool(cmd) {
-		return insts
+		return insts, nil
 	}
 
 	for _, inst := range instances {
 		// TODO: consider merging errors of multiple files, but ensure
 		// duplicates are removed.
-		exitOnErr(cmd, inst.Validate(), !flagIgnore.Bool(cmd))
+		err := inst.Validate()
+		if err != nil {
+			if flagIgnore.Bool(cmd) {
+				printError(cmd, err)
+			} else {
+				return nil, err
+			}
+		}
 	}
-	return insts
+	return insts, nil
 }
 
-func buildToolInstances(binst []*build.Instance) ([]*cue.Instance, error) {
-	instances := cue.Build(binst)
+func buildToolInstances(ctx *cue.Context, binst []*build.Instance) ([]*cue.Instance, error) {
+	// Reuse the same context, if there is one, so that the @embed interpreter can be used.
+	// Note that ctx may be nil when we do `cue help cmd`.
+	r := new(cue.Runtime)
+	if ctx != nil {
+		r = (*cue.Runtime)(ctx)
+	}
+	instances, _ := r.BuildInstances(binst)
 	for _, inst := range instances {
 		if inst.Err != nil {
 			return nil, inst.Err
@@ -784,8 +777,7 @@ func buildTools(cmd *Command, args []string) (*cue.Instance, error) {
 	}
 	loadCfg := *cfg.loadCfg
 	loadCfg.Tools = true
-	f := cmd.cmd.Flags()
-	setTags(&loadCfg, f, cmd.cmd.Parent().Flags())
+	setTags(&loadCfg, cmd.cmdCmd.Flags())
 
 	binst := loadFromArgs(args, &loadCfg)
 	if len(binst) == 0 {
@@ -794,6 +786,10 @@ func buildTools(cmd *Command, args []string) (*cue.Instance, error) {
 	included := map[string]bool{}
 
 	ti := binst[0].Context().NewInstance(binst[0].Root, nil)
+	// For @embed to also work in _tool.cue files, we must pass the module info along.
+	ti.Module = binst[0].Module
+	ti.Root = binst[0].Root
+
 	for _, inst := range binst {
 		k := 0
 		for _, f := range inst.Files {
@@ -810,7 +806,7 @@ func buildTools(cmd *Command, args []string) (*cue.Instance, error) {
 		inst.Files = inst.Files[:k]
 	}
 
-	insts, err := buildToolInstances(binst)
+	insts, err := buildToolInstances(cmd.ctx, binst)
 	if err != nil {
 		return nil, err
 	}
@@ -820,10 +816,11 @@ func buildTools(cmd *Command, args []string) (*cue.Instance, error) {
 		inst = cue.Merge(insts...)
 	}
 
-	r := value.ConvertToRuntime(inst.Value().Context())
+	ctx := inst.Value().Context()
 	for _, b := range binst {
 		for _, i := range b.Imports {
-			if _, err := r.Build(i); err != nil {
+			val := ctx.BuildInstance(i)
+			if err := val.Err(); err != nil {
 				return nil, err
 			}
 		}
